@@ -5,6 +5,6 @@ Created on Sun Dec 27 21:59:44 2020
 @author: Xingyan Liu
 """
 
-from .unaligned import datapair_from_adatas, DataPair
+from .unaligned import datapair_from_adatas, DataPair, make_features
 from .aligned import aligned_datapair_from_adatas, AlignedDataPair
 
