@@ -4,9 +4,11 @@ Created on Fri Oct 23 12:31:28 2020
 
 @author: Xingyan Liu
 """
-from typing import Union, Mapping, Sequence, Optional
+import logging
+from typing import Union, Mapping, Sequence, Optional, Callable
 from pathlib import Path
 
+import networkx as nx
 import numpy as np
 import pandas as pd
 from scipy import sparse
@@ -61,13 +63,14 @@ def view_color_map(cmap='viridis', n=None, figsize=(6, 2), s=150, k=20,
     
     Examples
     --------
-    import funx as fx
-    colors = ['Set1', 'viridis', 'Spectral']
-    fx.view_color_map(colors[-1], n=20)
-    
-    cmap = sc.pl.palettes.zeileis_26
-    cmap = sc.pl.palettes.default_64
-    fx.view_color_map(cmap, k=16)    
+
+    >>> colors = ['Set1', 'viridis', 'Spectral']
+    >>> view_color_map(colors[-1], n=20)
+
+    >>> import scanpy as sc
+    >>> cmap = sc.pl.palettes.zeileis_26
+    >>> cmap = sc.pl.palettes.default_64
+    >>> view_color_map(cmap, k=16)
     """
     if not isinstance(cmap, (np.ndarray, list)):
         #        from matplotlib import cm
@@ -113,42 +116,65 @@ def diy_cmap_grey_bg(name_fg='RdPu', low=0.15, rm_high=0.01, n=100):
 # Visualization of group correspondences on an alluvial plot
 
 
-def _alluvial_dict_from_confusdf(df: pd.DataFrame, vals_ignore=[0, np.nan]):
+def _alluvial_dict_from_confusdf(
+        df: pd.DataFrame, vals_ignore=(0, np.nan), tag=''):
+    """
+
+    Parameters
+    ----------
+    df:
+        might be the confusion matrix
+    vals_ignore
+    tag: str
+        used to make avoid name collisions in each side
+
+    Returns
+    -------
+
+    """
     dct = {}
     rowsum = df.sum(1)
 
     for i, idx in enumerate(df.index):
-        #        print(i, idx, type(rowsum[idx]), rowsum[idx])
+        # print(i, idx, type(rowsum[idx]), rowsum[idx])
         if rowsum[idx] in vals_ignore:
             continue
         dct[idx] = {}
         for col in df.columns:
             val = df.loc[idx, col]
             if val > 0:
-                dct[idx][col + ' '] = val  # make sure no collisions for names in each side
+                # avoid name collisions in each side
+                dct[idx][col + tag] = val
     return dct
 
 
-def alluvial_plot(confsdf: pd.DataFrame,
-                  labels=['', ''],  # ['true', 'predicted'],
-                  label_shift=0,  # -18
-                  title='alluvial plot',
-                  alpha=0.75,
-                  cmap_name='tab20_r',
-                  shuffle_colors=True,
-                  fonsize_title=11,
-                  figsize=(5, 5),
-                  fname=None,
-                  **kwds):
+def alluvial_plot(
+        confsdf: pd.DataFrame,
+        labels=['', ''],  # ['true', 'predicted'],
+        label_shift=0,  # -18
+        title='alluvial plot',
+        alpha=0.75,
+        cmap_name='tab20_r',
+        shuffle_colors=True,
+        fonsize_title=11,
+        figsize=(5, 5),
+        fname=None,
+        a_sort=None, b_sort=None,
+        **kwds):
     """ visualizing confusion matrix using alluvial plot
+
+    a_sort, b_sort: sorted group names on the left and the right side
     """
-    input_dct = _alluvial_dict_from_confusdf(confsdf)
+    input_dct = _alluvial_dict_from_confusdf(confsdf, tag=' ')
+    if b_sort is not None:
+        b_sort = [s + ' ' for s in b_sort]
     cmap = mpl.cm.get_cmap(cmap_name)
     ax = _alluvial.plot(input_dct, labels=labels,
                         label_shift=label_shift,
                         cmap=cmap, alpha=alpha,
                         shuffle_colors=shuffle_colors,
                         figsize=figsize,
+                        a_sort=a_sort, b_sort=b_sort,
                         **kwds)
     ax.set_title(title, fontsize=fonsize_title)
     fig = ax.get_figure()
@@ -158,10 +184,59 @@ def alluvial_plot(confsdf: pd.DataFrame,
     return ax
 
 
-# In[]
+def plot_stacked_bar(
+        df, norm=True, figsize=(6, 4),
+        colors=None,
+        cmap: Union[str, Callable] = 'tab20b',
+        legend_loc=(1.02, 0.01),
+):
+    """ helper function for visualizing the group compositions (e.g., in
+    each stage or condition).
+
+    Parameters
+    ----------
+    df: pd.DataFrame
+        the values in each row will be display as one stacked bar.
+    norm: bool
+        whether to normalize each row to unit-sum.
+    figsize: tuple
+    colors: Sequence
+    cmap: Union[str, Callable]
+        color-map function or color-map name
+    legend_loc: tuple
+        the legend location
+
+    Returns
+    -------
+    ax
+    """
+    if norm:
+        df = df.apply(lambda x: x / x.sum(), axis=1)
+
+    groups = list(df.columns)
+    x = list(df.index)
+    if colors is None:
+        if cmap is None:
+            cmap_func = plt.cm.get_cmap(cmap, len(groups))
+        elif hasattr(cmap, '__call__'):
+            cmap_func = cmap
+        else:
+            raise ValueError('cmap should be either Callable or str')
+        colors = [cmap_func(i) for i, _ in enumerate(groups)]
+
+    fig, ax = plt.subplots(figsize=figsize)
+    bott = df[groups[0]]
+    ax.bar(x, bott, label=groups[0], color=colors[0])
+    for i, c in enumerate(groups[1:]):
+        height = df[c].fillna(0.)
+        ax.bar(x, height, bottom=bott, label=c, color=colors[i + 1])
+        bott += height
+
+    ax.legend(loc=legend_loc)
+    return ax
+
+
 # functions for plotting confusion or contingency matrix
-
-
 def plot_contingency_mat(
         y_true, y_pred,
         norm_axis=1,
@@ -173,6 +248,7 @@ def plot_contingency_mat(
         fp: Union[Path, str, None] = None,
         **kwds
 ):
+    """function for plotting the contingency matrix"""
     from .analyze import wrapper_contingency_mat
     contmat = wrapper_contingency_mat(
         y_true, y_pred, normalize_axis=norm_axis,
@@ -193,8 +269,11 @@ def plot_confus_mat(y_true, y_pred, classes_on=None,
                     figsize: tuple = (4, 3),
                     ax=None, fp=None,
                     **kwargs):
-    """ by default, normalized by row (true classes)
+    """function for plotting the contingency matrix
+    by default, the values will be normalized by row (true classes)
     """
+    y_true = list(map(str, y_true))
+    y_pred = list(map(str, y_pred))
     if classes_on is None:
         classes_on = list(set(y_true).union(y_pred))
 
@@ -387,13 +466,40 @@ def heatmap(df_hmap: pd.DataFrame,
 
 def heatmap_probas(
         df_data, lbs,  # sort_lbs=False,
-        name_label=None,
+        name_label='Group',
         cmap_heat='magma_r', cmap_lb='tab20',
         figsize=(8, 4),
         vmax=1, vmin=0,
         xrotation=30,
         fp=None):
-    """ heatmap showing probabilities
+    """ Heatmap of the prediction probabilities
+
+    Parameters
+    ----------
+    df_data
+        A DataFrame of shape ``(n_samples, n_ref_classes)``.
+        Note that the samples should be sorted by the (query) group labels,
+        i.e., ``lbs``.
+    lbs
+        the (query) group labels
+    name_label
+        the name of the group labels
+    cmap_heat
+        the color-map of the heatmap
+    cmap_lb
+        the color-map of the group labels (``lbs``).
+    figsize
+        the figure size
+    vmax, vmin
+        the maximum and the minimum of the heatmap
+    xrotation
+        rotation of the x-tick-labels
+    fp
+        file-path for saving the plot
+
+    Returns
+    -------
+    gs (gridspec)
     """
     lbs = pd.Categorical(lbs, )
     pix_lbs = np.vstack([lbs.codes] * 2)
@@ -430,7 +536,8 @@ def heatmap_probas(
 
     # setting class-labels (x-ticks)
     classes = pd.unique(lbs)
-    cut_loc = np.hstack([[0], np.flatnonzero(np.diff(lbs.codes)), [len(lbs)]])  # // 2
+    cut_loc = np.hstack(
+        [[0], np.flatnonzero(np.diff(lbs.codes)), [len(lbs)]])
     x_loc = cut_loc[:-1] + np.diff(cut_loc) // 2
     ax2.set_xticks(x_loc)
     if xrotation:
@@ -559,23 +666,25 @@ def wrapper_heatmap_scores(
         fp=None,
         **kwds
 ):
-    """ sort columns and rows, plot heatmap of celltype scores """
+    """ sort columns and rows, plot heatmap of cell-type scores
+    """
     cols_anno = [col_label, col_pred]
     df_lbs = obs[cols_anno]
     if ignore_index:
         df_lbs = df_lbs.copy()
         df_lbs.index = df_score.index
+    # sort samples by labels
     df_lbs = df_lbs.sort_values(cols_anno)
     if n_subsample:
         from .base import subsample_each_group
         indices = subsample_each_group(df_lbs[col_label], n_out=n_subsample)
     else:
         indices = df_lbs.index
-    cols_ordered = [c for c in sorted(df_lbs[col_pred].unique())
+    cols_ordered = [c for c in sorted(set(df_lbs[col_pred]))
                     if c in df_score.columns]
-    
-    df_data = df_score.loc[indices][cols_ordered].copy()
-    lbs = df_lbs[col_label].loc[indices]
+
+    df_data = df_score.loc[indices, cols_ordered].copy()
+    lbs = df_lbs.loc[indices, col_label]
 
     gs = heatmap_probas(
         df_data.T, lbs, name_label=name_label,
@@ -584,12 +693,192 @@ def wrapper_heatmap_scores(
     return gs
 
 
+# scatter plots
+def sorted_scatter(
+        x, y, v=None, s=1.5,
+        title=None,
+        ax=None, cmap='RdYlBu_r',
+        marker='.',
+        vmin=None, vmax=None,
+        xlabel=None, ylabel=None,
+        with_cbar=True,
+        fontsize_cbar=9,
+        fontsize_axlabels=9,
+        fontsize_title=9,
+        **kwargs
+):
+    """ scatter plot colored with continuous values, with points of the higher
+    values on the top.
+
+    Parameters
+    ----------
+    x
+        the x coordinates
+    y
+        the y coordinates
+    v
+        the values for each point
+    s
+        the size of the scatters
+
+    Returns
+    -------
+    ax
+    """
+
+    if v is not None:
+        # points with high values will be put on the top
+        x, y, v = list(zip(*sorted(zip(x, y, v), key=lambda xyv: xyv[-1])))
+        x, y, v = np.array(x), np.array(y), np.array(v)
+
+    if ax is None:
+        fig, ax = plt.subplots()
+
+    cax = ax.scatter(
+        x, y, c=v, s=s,
+        cmap=cmap, marker=marker,
+        vmax=vmax, vmin=vmin, **kwargs)
+    ax.grid(False)
+    # remove y and x ticks
+    ax.set_yticks([])
+    ax.set_xticks([])
+    if xlabel is not None:
+        ax.set_xlabel(xlabel, fontsize=fontsize_axlabels)
+    if ylabel is not None:
+        ax.set_ylabel(ylabel, fontsize=fontsize_axlabels)
+    if title:
+        ax.set_title(title, fontsize=fontsize_title)
+
+    ax.set_aspect('equal', adjustable='datalim')
+    if with_cbar and v is not None:
+        cbar = plt.colorbar(cax, ax=ax, pad=0.01, fraction=0.08, aspect=30, )
+        for t in cbar.ax.get_yticklabels():
+            t.set_fontsize(fontsize_cbar)
+    return ax
+
+
+def embed_with_values(
+        xy: np.ndarray, values: Union[pd.Series, Mapping],
+        do_zscore: bool = False,
+        s: Union[int, None] = 1.5,
+        ncols=5, axscale=2.5,
+        name_xy='UMAP', with_cbar=True,
+        cmap='RdYlBu_r',
+        vmin=None, vmax=None,
+        **kwargs
+):
+    """ Visualize the given x-y coordinates, colored by the given (dict of) values
+
+    Parameters
+    ----------
+    xy: np.array
+        coordinates of each points, of shape (n_points, 2)
+    values: Mapping, pd.DataFrame or pd.Series
+        where the keys will be the title of subplots and each of the values
+        should be a vector of the length equal to xy.shape[0]
+    do_zscore
+        whether to calculate and display the z-scores instead of the original
+        values. (False)
+    s
+        the size of the scatters / points (1.5)
+    ncols
+        the number of columns for multiple-subplots (5)
+    axscale
+        adjust the size of the figures (2.5)
+    with_cbar
+        whether to display the color bars (True)
+    cmap
+        the colormap
+    vmin, vmax
+        the minimum and the maximum of the values
+    kwargs
+        Other Parameters for ``sorted_scatter``
+    """
+    x, y = xy[:, 0], xy[:, 1]
+    if isinstance(values, pd.Series):
+        values = {values.name: list(values)}
+    elif do_zscore:  # pd.DataFrame or Mapping[str, Sequence]
+        from sklearn.preprocessing import StandardScaler
+        keys = list(values.keys())
+        arr = np.array([list(values[k]) for k in keys])
+        values = pd.DataFrame(
+            StandardScaler(with_mean=True).fit_transform(arr).T, columns=keys)
+    cnames = values.keys()
+    n_plots = len(cnames)
+    if ncols >= n_plots:
+        nrows, ncols = 1, n_plots
+    else:
+        nrows = n_plots // ncols + min(n_plots % ncols, 1)
+    fig, axs = plt.subplots(
+        nrows, ncols, figsize=(ncols * axscale * 1.1, nrows * axscale),
+        subplot_kw=dict(aspect='equal'))
+
+    xlabel = f'{name_xy}1'
+    ylabel = f'{name_xy}2'
+    if hasattr(axs, 'flatten'):
+        axs_flatten = axs.flatten()
+    else:
+        axs_flatten = np.array([axs])
+    s = 6000 / len(x) if s is None else s
+    for ax, cname in zip(axs_flatten, cnames):
+        v = list(values[cname])
+        sorted_scatter(x, y, v, s=s, ax=ax,
+                       title=cname,
+                       cmap=cmap,
+                       vmin=vmin, vmax=vmax,
+                       with_cbar=with_cbar,
+                       xlabel=xlabel, ylabel=ylabel, **kwargs)
+    for ax in axs_flatten[n_plots:]:
+        ax.set_frame_on(False)
+        ax.grid(False)
+        # remove y and x ticks
+        ax.set_yticks([])
+        ax.set_xticks([])
+
+    return fig, axs
+
+
+def adata_embed_with_values(
+        adata: sc.AnnData, values: Union[pd.Series, Mapping],
+        embed_key='UMAP', fp=None, figsize_save=None, **kwargs):
+    """
+    This function can also be used to visualize the average expressions of some
+    cell types on gene embedding, in which case, each observation in
+    `adata` represents a genes.
+
+    Parameters
+    ----------
+    adata
+        the ``AnnData`` object
+    values: Mapping, pd.DataFrame or pd.Series
+        where the keys will be the title of subplots and the values
+        should be of the length equal to adata.shape[0]
+    embed_key
+        the type of the embeddings, e.g., "UMAP" or "TSNE"
+    fp: Path or str
+        path to save figure
+    figsize_save: (int, int)
+        the figure size of the saved plot.
+    Returns
+    -------
+    fig, axs
+    """
+    obsm_embed_key = 'X_{}'.format(embed_key.lower())
+    xy = adata.obsm[obsm_embed_key]  # n_pts x 2
+    fig, axs = embed_with_values(xy, values, name_xy=embed_key, **kwargs)
+    if fp:
+        _save_with_adjust(fig, fp, figsize=figsize_save)
+    fig.show()
+    return fig, axs
+
+
 # In[]
 def _get_affine_mat(angle_x=30, angle_y=150):
     """ rotate x and y axis to mock 3D projection
     if angle_x=0, angle_y=90, an identity matrix will be returned.
     """
-    angle_x, angle_y = tuple(map(lambda x: (x / 180) * np.pi, (angle_x, angle_y)))
+    angle_x, angle_y = tuple(
+        map(lambda x: (x / 180) * np.pi, (angle_x, angle_y)))
     _transmat = np.array([[np.cos(angle_x), np.sin(angle_x)],
                           [np.cos(angle_y), np.sin(angle_y)]
                           ])
@@ -665,10 +954,12 @@ def plot_mapped_graph(
         plot_edges_by_adj(adj1, xy1, ax=ax, zorder=0.9, **_kwds_edge)
     if adj2 is not None:
         plot_edges_by_adj(adj2, xy2, ax=ax, zorder=0.9, **_kwds_edge)
-    plot_edges_by_adj(mapping, np.vstack([xy1, xy2]), ax=ax, zorder=1, **_kwds_edge)
+    plot_edges_by_adj(mapping, np.vstack([xy1, xy2]), ax=ax, zorder=1,
+                      **_kwds_edge)
 
     cmap_pt = cmap_pt if isinstance(cmap_pt, (list, tuple)) else [cmap_pt] * 2
-    ax.scatter(x1, y1, marker='.', c=pt_color1, s=pt_size, cmap=cmap_pt[0], zorder=2.5)
+    ax.scatter(x1, y1, marker='.', c=pt_color1, s=pt_size, cmap=cmap_pt[0],
+               zorder=2.5)
     ax.scatter(x2, y2, marker='.', c=pt_color2, s=pt_size, cmap=cmap_pt[1])
     ax.set_axis_off()
 
@@ -721,12 +1012,13 @@ def embedding_mock3d(
     else:
         pt_color = None
     if ax is None:
-        fig, ax = plt.subplots(figsize=figsize) # w*h
+        fig, ax = plt.subplots(figsize=figsize)  # w*h
     if edges:
         nn_key = f'{neighbor_key}_connectivities' if neighbor_key else 'connectivities'
         adj = adata.obsp[nn_key]
         plot_edges_by_adj(adj, xy, ax=ax, zorder=0.9, **kwds_edge)
-    ax.scatter(x, y, marker='.', c=pt_color, s=pt_size, cmap=cmap_pt, zorder=2.5)
+    ax.scatter(x, y, marker='.', c=pt_color, s=pt_size, cmap=cmap_pt,
+               zorder=2.5)
     ax.set_axis_off()
     if fp: _save_with_adjust(ax.figure, fp, dpi=400)
     return ax
@@ -842,13 +1134,17 @@ def _get_value_index_list(lst, vals, return_vals=False):
     lst = list(lst)
     ids = []
     _vals = []
+    missed = []
     for v in vals:
         try:
             i = lst.index(v)
             ids.append(i)
             _vals.append(v)
         except ValueError:
+            missed.append(v)
             continue
+    if len(missed) >= 1:
+        logging.warning(f"{missed} is/are not in the list")
     if return_vals:
         return ids, _vals
     return ids
@@ -865,6 +1161,37 @@ def umap_with_annotates(
         ax=None,
         fp=None,
         **plkwds):
+    """ Plot the UMAP embeddings and annotate the names of the given points
+
+    Parameters
+    ----------
+    adt
+    text_ids
+        ids of the points to be annotated, corresponding to
+        ``anno_df.index`` or ``anno_df[index_col]`` (if ``index_col`` is given)
+    color
+        a column name in ``adt.obs``
+    anno_df
+        use ``adt.obs`` as the default annotations.
+    text_col
+        used as the annotation texts.
+        if specifies, should be a column name in ``anno_df``.
+    index_col
+        used for index the points, corresponding to ``text_ids``
+        if specifies, should be a column name in ``anno_df``.
+    anno_fontsize
+        the fontsize of the annotations
+    ax
+        the plot ax
+    fp
+        file path for saving the plot.
+    plkwds
+        other key-word-args for ``sc.pl.umap``
+
+    Returns
+    -------
+    ax
+    """
     from adjustText import adjust_text
     anno_df = adt.obs if anno_df is None else anno_df
 
@@ -873,19 +1200,18 @@ def umap_with_annotates(
         index_all, text_ids, return_vals=True)  # only the first one
     # subsetting xumap and df
     xumap = adt.obsm['X_umap'][idnums, :]
-    #    anno_df = anno_df.iloc[idnums, :]
     xs = xumap[:, 0]
     ys = xumap[:, 1]
     texts = _text_ids if text_col is None else anno_df[text_col][idnums]
-    #    print(texts)
 
     if ax is None:
         fig, ax = plt.subplots()
     sc.pl.umap(adt, color=color, ax=ax, show=False, **plkwds)
     # annotation
-    _texts = [ax.text(x, y, t, {'fontsize': anno_fontsize}) for x, y, t in zip(xs, ys, texts)]
+    _texts = [ax.text(x, y, t, {'fontsize': anno_fontsize}) for x, y, t in
+              zip(xs, ys, texts)]
     adjust_text(_texts, ax=ax, arrowprops=dict(arrowstyle='-', color='k'))
-    _save_with_adjust(fig, fp)
+    _save_with_adjust(ax.figure, fp)
     return ax
 
 
@@ -946,12 +1272,10 @@ def _adjust_ylims(ax, scale=1.):
 
 def plot_multipartite_graph(
         g,
-        node_list=None,
+        # node_list=None,
         subset_key='subset',
         weight_key='weight',
         nodelb_key='original name',
-        nodesz_key='size',
-        node_color='lightblue',
         alpha=0.85,
         figsize=(10, 8), fp=None,
         xscale=1.5, yscale=1.1,
@@ -960,13 +1284,19 @@ def plot_multipartite_graph(
         node_size_min=100,
         node_size_max=1800,
         with_labels=True,
+        colors=None,
         **kwds):
+    """display the abstracted multipartite graph
+    (cellType - geneModules - geneModules - cellTypes)
+
+    """
     import networkx as nx
-    if node_list is None: node_list = list(g)
+    # if node_list is None: node_list = list(g)
 
     pos = multipartite_layout(g, subset_key=subset_key, )
 
-    edge_vals = [v * edge_scale for v in nx.get_edge_attributes(g, weight_key).values()]
+    edge_vals = [v * edge_scale for v in
+                 nx.get_edge_attributes(g, weight_key).values()]
     fig, ax = plt.subplots(figsize=figsize)
 
     ax.set_axis_off()
@@ -975,10 +1305,10 @@ def plot_multipartite_graph(
     nx.draw_networkx_edges(
         g, pos,
         width=edge_vals,
-        #            edge_color='grey',
+        # edge_color='grey',
         edge_color=edge_vals, edge_cmap=_cut_cmap('Greys', ),
         alpha=alpha, )
-    nodedf = _prepare_for_nxplot(g, **kwds)
+    nodedf = _prepare_for_nxplot(g, colors=colors, **kwds)
 
     #    node_sizes = np.array([nx.get_node_attributes(g, nodesz_key)[nd] for nd in node_list])
     node_sizes = np.maximum(nodedf['plt_size'] * node_scale, node_size_min)
@@ -992,8 +1322,6 @@ def plot_multipartite_graph(
     #    ax = plot_graph(g, pos)
     if xscale is not None:
         _adjust_xlims(ax, xscale)
-    #        xlims = ax.get_xlim()
-    #        ax.set_xlim((xlims[0] * xscale, xlims[1] * xscale))
     if yscale is not None:
         _adjust_ylims(ax, yscale)
 
@@ -1004,10 +1332,12 @@ def plot_multipartite_graph(
 def _prepare_for_nxplot(
         g, names=('cell group', 'gene module'),
         sizes=(1800, 900, 900, 1800),
-        colors=('pink', 'lightblue', 'lightblue', 'pink')):
+        colors=None):
     """
     g: a multipartite graph with 4 layers
     """
+    if colors is None:
+        colors = ('pink', 'lightblue', 'lightblue', 'pink')
     nodedf = pd.DataFrame(g.nodes.values(), index=g.nodes.keys())
     nodedf['ntype'] = nodedf['subset'].apply(
         lambda x: names[0] if x in [0, 3] else names[1]
@@ -1015,7 +1345,8 @@ def _prepare_for_nxplot(
     nodedf['plt_color'] = nodedf['subset'].apply(
         lambda x: colors[x]
     )
-    nodedf['plt_size'] = nodedf.groupby('ntype')['size'].apply(lambda x: x / x.max())
+    nodedf['plt_size'] = nodedf.groupby('ntype')['size'].apply(
+        lambda x: x / x.max())
 
     for isubset, size_scale in enumerate(sizes):
         inds_sub = nodedf['subset'] == isubset
@@ -1116,7 +1447,7 @@ def multipartite_layout(
         layers[layer] = [v] + layers.get(layer, [])
 
     nodes_per_layer = [len(item[1]) for item in layers.items()]
-    print(nodes_per_layer)
+    logging.info("nodes per layer: {nodes_per_layer}")
     max_nodes = max(nodes_per_layer)
     thd = max_nodes * int(spread_nodes)
     pos = None
