@@ -147,7 +147,8 @@ def adata_from_raw(dirname, backup_npz=True, name_mtx='matrix',
     elif f'{name_mtx}.mtx' in files:
         mat = load_sparse(dirname / f'{name_mtx}.mtx', backup_npz=backup_npz)
     else:
-        raise FileNotFoundError(f"None of file named `{name_mtx}.npz` or `{name_mtx}.mtx` exists!")
+        raise FileNotFoundError(
+            f"None of file named `{name_mtx}.npz` or `{name_mtx}.mtx` exists!")
     mat = sparse.csr_matrix(mat.T)
 
     params = dict(header=None, sep='\t', index_col=0)
@@ -173,9 +174,10 @@ def add_columns(
         copy=True, **kwannos):
     def _set_values(_df, k, v, ignore_index=ignore_index):
         if k in _df.columns:
-            print(f'NOTE that column "{k}" will be covered by new values')
+            logging.warning(
+                f'NOTE that column "{k}" will be covered by new values')
         _df[k] = list(v) if ignore_index else v
-        print(k, end=', ')
+        # print(k, end=', ')
 
     if copy: df0 = df0.copy()
     if df is not None:
@@ -190,7 +192,7 @@ def add_columns(
         for k, v in kwannos.items():
             # be careful of the index mismatching problems!!!
             _set_values(df0, k, v)
-    print('done!')
+    # print('done!')
     return df0 if copy else None
 
 
@@ -256,9 +258,12 @@ def make_adata(mat: Union[np.ndarray, sparse.spmatrix],
     return adata
 
 
-def merge_metas(adatas, obs_keys):
-    obs_keys = [obs_keys] if isinstance(obs_keys, str) else obs_keys
-    obss = [adt.obs[obs_keys] for adt in adatas]
+def merge_metas(adatas, obs_keys=None):
+    if obs_keys:
+        obs_keys = [obs_keys] if isinstance(obs_keys, str) else obs_keys
+        obss = [adt.obs[obs_keys] for adt in adatas]
+    else:
+        obss = [adt.obs for adt in adatas]
     return pd.concat(obss, axis=0)
 
 
@@ -314,16 +319,16 @@ def merge_adatas(
     mats, genes = merge_named_matrices(mats, genes, union=union)
     merged_mat = sparse.vstack(mats, )  # dtype = mats[0].dtype)
     genes = pd.DataFrame(index=genes)
-    print(genes.head())
+    logging.debug(genes.head())
 
-    if obs_keys is None:
-        obs = pd.DataFrame(index=np.concatenate(bcds, axis=0))
-    else:
-        print("merging metadata...")
-        obs = merge_metas(_adatas, obs_keys)
+    # if obs_keys is None:
+    #     obs = pd.DataFrame(index=np.concatenate(bcds, axis=0))
+    # else:
+    logging.info("merging metadata...")
+    obs = merge_metas(_adatas, obs_keys)
     if key_dsname is not None:
         obs[key_dsname] = pd.Categorical(lbs_datasets, categories=dsnames)
-    print(obs.head())
+    logging.debug(obs.head())
     return make_adata(merged_mat, obs=obs, var=genes)
 
 
@@ -379,7 +384,8 @@ def merge_named_matrices(
         if not union and (dsnames is not None) and verbose:
             print('After {}: {} genes'.format(dsnames[idx], len(keep_genes)))
         if len(keep_genes) == 0:
-            raise ValueError('Error: No genes found in all datasets, exiting...')
+            raise ValueError(
+                'Error: No genes found in all datasets, exiting...')
     if verbose:
         print('Found {} genes among all datasets'
               .format(len(keep_genes)))
@@ -396,7 +402,8 @@ def merge_named_matrices(
             gene_to_idx = {gene: idx for idx, gene in enumerate(gene_lists[i])}
             for j, gene in enumerate(union_genes):
                 if gene in gene_to_idx:
-                    X_new[:, j] = X_old[:, gene_to_idx[gene]].toarray().flatten()
+                    X_new[:, j] = X_old[:,
+                                  gene_to_idx[gene]].toarray().flatten()
             mat_list[i] = sparse.csr_matrix(X_new)
         # print('TEST:', mat_list[i].data[:5])
         ret_genes = np.array(union_genes)
@@ -448,7 +455,8 @@ def align_adata_vars(adata1: sc.AnnData,  # better be raw data
     """
     vars_all1, vars_all2 = list(map(all_vars_of_adata, [adata1, adata2]))
     if df_varmap_1v1 is None:
-        vars1 = vars2 = list(set(adata1.var_names).intersection(adata2.var_names))
+        vars1 = vars2 = list(
+            set(adata1.var_names).intersection(adata2.var_names))
     else:
         submaps_1v1 = subset_matches(df_varmap_1v1, vars_all1, vars_all2,
                                      union=False)
@@ -495,9 +503,10 @@ def change_names(names: Sequence,
     return list(map(foo_change, names, **kwargs))
 
 
-def get_homologues(df_match: pd.DataFrame,
+def get_homologies(df_match: pd.DataFrame,
                    vals: Sequence,
-                   cols: Union[None, Sequence[str]] = None,  # [col_from, col_to]
+                   cols: Union[None, Sequence[str]] = None,
+                   # [col_from, col_to]
                    reverse: bool = False,
                    uniquelist: bool = True,
                    with_null=False):
@@ -544,7 +553,7 @@ def get_homologues(df_match: pd.DataFrame,
     homos = pd.concat(homos, axis=0, ignore_index=True)
     # homos = _df_match.loc[vals, col_to]
     if len(null) >= 1:
-        print(f'the following have no homologies:\n{null}')
+        logging.warning(f'the following have no homologies:\n{null}')
     if uniquelist:
         homos = pd.unique(homos[col_to].values)
         return list(homos) + null if with_null else homos
@@ -552,6 +561,9 @@ def get_homologues(df_match: pd.DataFrame,
         return (homos, null) if with_null else homos
 
 
+get_homologues = get_homologies
+
+
 def subset_matches(df_match: pd.DataFrame,
                    left: Sequence,
                    right: Sequence,
@@ -732,7 +744,7 @@ def make_bipartite_adj(df_map: pd.DataFrame,
     return bi_adj, nodes1, nodes2
 
 
-def pivot_df_to_sparse(df, row=0, col=1, key_data=None, **kwds):
+def pivot_df_to_sparse(df: pd.DataFrame, row=0, col=1, key_data=None, **kwds):
     """
     row, col:
         str or int, int for column index, and str for column name
@@ -791,7 +803,7 @@ def pivot_to_sparse(rows: Sequence, cols: Sequence,
         if rownames is not None:
             tmp = set(rownames)
             r_kept = list(map(lambda x: x in tmp, rows))
-            print(sum(r_kept))
+            logging.debug(sum(r_kept))
         else:
             r_kept = np.ones_like(rows).astype(bool)
         if colnames is not None:
@@ -800,7 +812,7 @@ def pivot_to_sparse(rows: Sequence, cols: Sequence,
         else:
             c_kept = np.ones_like(rows).astype(bool)
         kept = np.minimum(r_kept, c_kept)
-        print(len(kept), sum(kept))
+        logging.debug(len(kept), sum(kept))
         rows, cols, data = rows[kept], cols[kept], data[kept]
 
     (rownames, ii), (colnames, jj) = list(map(
@@ -811,6 +823,66 @@ def pivot_to_sparse(rows: Sequence, cols: Sequence,
     return sparse_mat, rownames, colnames
 
 
+class AdjacentTrans(object):
+    def __init__(self, adj, vars1, vars2, trans_to=0):
+        assert adj.shape[0] == len(vars1)
+        assert adj.shape[1] == len(vars2)
+        self.adj = adj
+        self.vars1 = vars1  # for rows
+        self.vars2 = vars2  # for columns
+        self.trans_to = int(trans_to)  # 0 for `vars1` and 1 for `vars2`
+
+    @property
+    def shape(self):
+        return self.adj.shape
+
+    @staticmethod
+    def from_edge_df(edge_df, col_weight=None):
+        # edge_df: the first column for rows and the second for columns
+        adj, vars1, vars2 = pivot_df_to_sparse(edge_df, key_data=col_weight)
+        return AdjacentTrans(adj, vars1, vars2)
+
+    def reduce_to_align(self, sep='__'):
+        trans_to = self.trans_to
+        logging.info(f"[AdjacentTrans] reduce to align, trans_to={trans_to}")
+        if trans_to == 0:
+            return self.vars1, reduce_to_align(self.adj, self.vars2, sep=sep)
+        else:
+            return reduce_to_align(self.adj.T, self.vars1, sep=sep), self.vars2
+
+    def reduce_to_align_features(self, feats):
+        """feats: a sample-by-vars matrix or (sparse) array"""
+        trans_to = self.trans_to
+        adj = self.adj
+        logging.info(
+            f"[AdjacentTrans] reduce features to align, trans_to={trans_to}")
+        if trans_to == 1:
+            adj = adj.T
+        assert feats.shape[1] == adj.shape[1]
+        feats = adj.dot(feats.T) / adj.sum(1)  # divide row-sums as averages
+        # if self.trans_to == 0:
+        #     assert feats.shape[1] == adj.shape[1]
+        #     feats = adj.dot(feats.T) / adj.sum(1)
+        # else:
+        #     adj = adj.T
+        #     assert feats.shape[1] == adj.shape[1]
+        #     feats = adj.T.dot(feats.T)
+        return feats.T
+
+    def T(self):
+        """Transpose"""
+        return AdjacentTrans(self.adj.T, self.vars2, self.vars1)
+
+
+def reduce_to_align(adj: sparse.spmatrix, names: Sequence, sep='__'):
+    """reduce column-names to align rows"""
+    assert adj.shape[1] == len(names)
+    adj = sparse.csr_matrix(adj)
+    # adj.indices.shape == adj.data.shape
+    return [sep.join(np.take(names, adj[i, :].indices).astype(str))
+            for i in range(adj.shape[0])]
+
+
 def label_binarize_each(labels, classes, sparse_out=True):
     lb1hot = label_binarize(labels, classes=classes, sparse_output=sparse_out)
     if len(classes) == 2:
@@ -823,7 +895,7 @@ def label_binarize_each(labels, classes, sparse_out=True):
 
 def agg_group_edges(adj, labels1, labels2=None,
                     groups1=None, groups2=None,
-                    asdf=True, verbose=True):
+                    asdf=True, verbose=False):
     """
     Parameters
     ----------
@@ -920,10 +992,10 @@ def _order_contingency_array(mat, axis=1):
     """
     order = np.argsort(np.argmax(mat, axis=axis))
     if axis == 1:
-        print('Re-order the rows')
+        logging.debug('Re-order the rows')
         return mat[order, :]
     else:
-        print('Re-order the columns')
+        logging.debug('Re-order the columns')
         return mat[:, order]
 
 
@@ -934,10 +1006,10 @@ def _order_contingency_df(df: pd.DataFrame, axis=1):
     """
     order = np.argsort(np.argmax(df.values, axis=axis))
     if axis == 1:
-        print('Re-order the rows')
+        logging.debug('Re-order the rows')
         return df.iloc[order, :]
     else:
-        print('Re-order the columns')
+        logging.debug('Re-order the columns')
         return df.iloc[:, order]
 
 
@@ -1075,7 +1147,8 @@ def take_adata_groups(adata: sc.AnnData,
     indicators = take_group_labels(adata.obs[key], group_names,
                                    indicate=True)
     if copy:
-        return adata[indicators, :].X.copy() if onlyx else adata[indicators, :].copy()
+        return adata[indicators, :].X.copy() if onlyx else adata[indicators,
+                                                           :].copy()
     else:
         return adata[indicators, :].X if onlyx else adata[indicators, :]
 
@@ -1238,7 +1311,7 @@ def _filter_mito(lst):
 
 
 def normalize_default(adata: sc.AnnData,
-                      target_sum=1e4,
+                      target_sum=None,
                       copy: bool = False,
                       log_only: bool = False,
                       force_return: bool = False, ):
@@ -1272,7 +1345,8 @@ def normalize_default(adata: sc.AnnData,
         logging.info('No copy was made, the input AnnData will be changed!')
     logging.info('normalizing datasets with default settings.')
     if not log_only:
-        logging.info(f'performing total-sum normalization, target_sum={target_sum}...')
+        logging.info(
+            f'performing total-sum normalization, target_sum={target_sum}...')
         sc.pp.normalize_total(adata, target_sum=target_sum)
     else:
         logging.info('skipping total-sum normalization')
@@ -1291,7 +1365,7 @@ def normalize_log_then_total(
         adata = adata.copy()
     logging.info(
         'normalizing datasets with default settings (log1p first).'
-        f'target_sum = {target_sum:.1f}')
+        f'target_sum = {target_sum}')
     sc.pp.log1p(adata)  # x <- log(x + 1)
     sc.pp.normalize_total(adata, target_sum=target_sum)
     return adata if copy or force_return else None
@@ -1395,7 +1469,8 @@ def group_zscore_adata(adt: sc.AnnData,
     """
     labels = adt.obs[groupby]
     if key == 'counts':
-        logging.info('Z-score scaling on count matrix, transformed into a dense array')
+        logging.info(
+            'Z-score scaling on count matrix, transformed into a dense array')
         if sparse.issparse(adt.X):
             X = adt.X.toarray()
         else:
@@ -1521,7 +1596,7 @@ def normalize_row(X, scale_factor=1, by='sum'):
         X_new = np.diag(norm_).dot(X)
 
     if isinstance(X, pd.DataFrame):
-        X_new.columns = X.columns
+        X_new = pd.DataFrame(X_new, columns=X.columns)
     return X_new
 
 
@@ -1635,10 +1710,47 @@ def group_value_counts(df, count_on, group_by, split=True, **kwds):
     return vcnt
 
 
-def group_mean(X: Union[np.ndarray, sparse.spmatrix],
-               labels: Sequence,
-               binary=False, classes=None, features=None,
-               print_groups=True):
+def group_mean(
+        X: Union[np.ndarray, sparse.spmatrix],
+        labels: Sequence,
+        binary: bool = False,
+        classes=None, features=None,
+        **kwds
+):
+    """ compute the group averaged features
+
+    Parameters
+    ----------
+    X: np.ndarray or sparse.spmatrix
+        shape (n_samples, n_features)
+    labels:
+        shape (n_samples, )
+    binary
+        if True, the results will be the non-zero proportions
+    classes:
+        optional, names of groups
+    features:
+        optional, names of features
+
+    Returns
+    -------
+    average_mat: pd.DataFrame
+    """
+    if sparse.issparse(X):
+        return group_mean_sparse(
+            X, labels, binary=binary, classes=classes, features=features, **kwds
+        )
+    else:
+        return group_mean_dense(
+            X, labels, binary=binary, classes=classes, features=features, **kwds
+        )
+
+
+def group_mean_sparse(
+        X: sparse.spmatrix,
+        labels: Sequence,
+        binary=False, classes=None, features=None,
+        print_groups=False):
     """
     This function may work with more efficiency than `df.groupby().mean()` 
     when handling sparse matrix.
@@ -1658,13 +1770,13 @@ def group_mean(X: Union[np.ndarray, sparse.spmatrix],
     classes = np.unique(labels, ) if classes is None else classes
     if binary:
         X = (X > 0)  # .astype('float')
-        print('Binarized...the results will be the expression proportions.')
+        logging.info('Binarized; the results will be non-zero proportions')
 
     if len(classes) == 1:
         grp_mean = X.mean(axis=0).T
     else:
         lb1hot = label_binarize_each(labels, classes=classes, sparse_out=True)
-        print(f'Calculating feature averages for {len(classes)} groups')
+        logging.info(f'Calculating feature averages for {len(classes)} groups')
         if print_groups:
             print(classes)
         grp_mean = X.T.dot(lb1hot) / lb1hot.sum(axis=0)
@@ -1676,17 +1788,22 @@ def group_mean_dense(
         X, labels, binary=False,
         index_name='group',
         classes=None,
+        features=None,
 ):
     classes = np.unique(labels, ) if classes is None else classes
     if binary:
         X = (X > 0)  # .astype('float')
-        logging.info('Binarized...the results will be the expression '
+        logging.info('Binarized...the results will be the non-zero '
                      'proportions.')
-    tmp = pd.DataFrame(X)
+    tmp = pd.DataFrame(X, columns=features)
     tmp[index_name] = list(labels)
-    avgs = tmp.groupby(index_name).mean()
+    avgs = tmp.groupby(index_name).mean().T
+    del tmp[index_name]
     # print(avgs.shape)
-    return avgs.T[classes]  # each column as a group
+    for c in classes:
+        if c not in avgs.columns:
+            avgs[c] = 0.
+    return avgs[classes]  # each column as a group
 
 
 def group_median_dense(
@@ -1760,8 +1877,9 @@ def group_mean_multiadata(adatas: Sequence[sc.AnnData],
         keys = [keys] * n
     tags = list(map(str, np.arange(n))) if tags is None else tags
     # within-group averages 
-    gmeans = [group_mean_adata(adt, ky, use_genes, binary=binary, use_raw=use_raw) \
-              for adt, ky in zip(adatas, keys)]
+    gmeans = [
+        group_mean_adata(adt, ky, use_genes, binary=binary, use_raw=use_raw) \
+        for adt, ky in zip(adatas, keys)]
     gmeans = pd.concat(gmeans, axis=1, keys=tags, sort=False)
     gmeans.fillna(0, inplace=True)
     # Make names
@@ -1826,7 +1944,7 @@ def quick_preprocess(
 
 def quick_pre_vis(adata, hvgs=None,
                   normalize_data=True,
-                  target_sum=1e4,
+                  target_sum=None,
                   batch_key=None,
                   n_pcs=30, nneigh=10,
                   vis='umap',
@@ -1902,6 +2020,7 @@ def get_scnet(adata: sc.AnnData):
 
 def get_hvgs(adata, force_redo=False, batch_key=None,
              n_top_genes=2000,
+             key_hvg: str = 'highly_variable',
              **hvg_kwds):
     """ packed function for computing and get HVGs from ``sc.AnnData`` object.
     if `force_redo` is True, data should be normalized and log-transformed.
@@ -1918,10 +2037,12 @@ def get_hvgs(adata, force_redo=False, batch_key=None,
     n_top_genes:
         Number of highly-variable genes to keep. 
         If specified, other parameters will be ignored.
+    key_hvg:
+        A column that should be in ``adata.var.columns``
     hvg_kwds:
         Other parameters, e.g. min_mean, max_mean, min_disp, max_disp
     """
-    key_hvg = 'highly_variable'
+    # key_hvg = 'highly_variable'
     if force_redo or key_hvg not in adata.var.columns:
         logging.info(
             f'performing HVG-selection...\n '
@@ -1935,22 +2056,42 @@ def get_hvgs(adata, force_redo=False, batch_key=None,
     return list(all_vars[is_hvg])
 
 
-def get_marker_info_table(adata, groups=None,
-                          key='rank_genes_groups'):
+def get_marker_info_table(
+        adata, groups=None, key='rank_genes_groups',
+        cut_padj: Optional[float] = 0.05,
+        cut_logfc: Optional[float] = 0.25,
+        cut_pts: Optional[float] = None,
+):
     result = adata.uns[key]
     if groups is None:
         groups = result['names'].dtype.names
 
     dfs = []
-    cols = ['names', 'logfoldchanges', 'pvals', 'pvals_adj', 'scores']
+    cols = ['names', 'logfoldchanges', 'pvals', 'pvals_adj', 'scores', ]
+    # cols = [c for c in cols if c in result.keys()]
+    flag_pts = 'pts' in result.keys()
+
     for group in groups:
         _df = pd.DataFrame({
             key: result[key][group] for key in cols
         })
         _df['group'] = group
-        dfs.append(_df[['group'] + cols])
-    df = pd.concat(dfs, axis=0)
-    return df
+
+        if flag_pts:
+            # expression proportions, avoid row mismatching
+            _df['pts'] = _df['names'].map(result['pts'][group])
+            _df['pts_rest'] = _df['names'].map(result['pts_rest'][group])
+            if cut_pts is not None:
+                _df = _df[_df['pts'] >= cut_pts]
+        if cut_padj is not None:
+            _df = _df[_df['pvals_adj'] <= cut_padj]
+        if cut_logfc is not None:
+            _df = _df[_df['logfoldchanges'] >= cut_logfc]
+        dfs.append(_df.copy())  # [['group'] + cols])
+    df = pd.concat(dfs, axis=0, keys=groups)
+    if flag_pts:
+        cols += ['pts', 'pts_rest']
+    return df[['group'] + cols]
 
 
 def get_marker_name_table(adata, key='rank_genes_groups'):
@@ -1984,12 +2125,49 @@ def top_markers_from_df(marker_df, n=5, groups=None, unique=True, ):
     return top
 
 
+def top_markers_from_info(
+        df_info, n=5, groups=None, unique=True,
+        col_group='group',
+        col_name='names') -> list:
+    """
+    df_info: DEG-info table that can be take from `top_markers_from_adata`
+    """
+    # if groups is not None:
+    #     df_info = df_info[df_info[col_group].isin(groups)]
+    subdf = df_info.groupby(col_group).apply(lambda x: x.head(n))
+    if groups is not None:
+        subdf = subdf.loc[groups]  # filter, or keep the group orders
+    names = subdf[col_name]
+    return names.unique().tolist() if unique else names.tolist()
+
+
 def top_markers_from_adata(adata: sc.AnnData,
-                           n=5, groups=None,
+                           n: Optional[int] = 5,
+                           groups: Optional[Sequence] = None,
                            unique=True,
+                           cut_padj=0.05,
+                           cut_logfc=0.25,
+                           cut_pts=0.1,
                            key='rank_genes_groups'):
-    df = get_marker_name_table(adata, key=key)
-    return top_markers_from_df(df, n=n, groups=groups, unique=unique)
+    df_info = get_marker_info_table(
+        adata, groups, key=key,
+        cut_padj=cut_padj, cut_logfc=cut_logfc, cut_pts=cut_pts)
+
+    if n is None:
+        if unique:
+            return df_info['names'].unique().tolist()
+        return df_info['names'].tolist()
+    else:
+        # df = get_marker_name_table(adata, key=key)
+        # return top_markers_from_df(df, n=n, groups=groups, unique=unique)
+        if groups is None:
+            _groupby = adata.uns[key]['params']['groupby']
+            try:
+                # to keep the group order
+                groups = adata.obs[_groupby].cat.categories
+            except:
+                pass
+        return top_markers_from_info(df_info, n=n, groups=groups, unique=unique)
 
 
 def compute_and_get_DEGs(adata: sc.AnnData,
@@ -2002,20 +2180,26 @@ def compute_and_get_DEGs(adata: sc.AnnData,
                          inplace=True,
                          do_normalize=False,
                          method='t-test',
+                         return_info=False,
+                         cuts={},
                          **kwds):
     """ Compute and get DEGs from ``sc.AnnData`` object
     
     By default, assume that the counts in adata has been normalized.
     If `force_redo`: re-compute DEGs and the original DEGs in adata will be ignored
+
+    cuts: dict with keys 'cut_padj', 'cut_pts', and 'cut_logfc'
+
     """
     if not inplace:
         logging.info('making a copy')
         adata = adata.copy()
-
+    adata.obs[groupby] = adata.obs[groupby].astype('category')
     if force_redo or key_added not in adata.uns.keys():
         logging.info(f'computing differentially expressed genes using {method}')
         if do_normalize:
-            normalize_default(adata, target_sum=1e4, )
+            # todo: ``target_sum`` used to be 1e4
+            normalize_default(adata, target_sum=None, )
         else:
             logging.info(
                 'computing differential expression genes using default settings'
@@ -2024,12 +2208,15 @@ def compute_and_get_DEGs(adata: sc.AnnData,
             adata = remove_adata_small_groups(adata, key=groupby, min_samples=1)
         sc.tl.rank_genes_groups(adata, groupby=groupby,
                                 key_added=key_added,
+                                pts=True,
                                 method=method,
                                 **kwds)
+    if return_info:
+        return get_marker_info_table(adata, key=key_added, **cuts)
     return top_markers_from_adata(adata, n=n,
                                   groups=groups,
                                   unique=unique,
-                                  key=key_added)
+                                  key=key_added, **cuts)
 
 
 @dec_timewrapper('leiden')
@@ -2058,3 +2245,127 @@ def get_leiden_labels(adata, hvgs=None,
     lbs = adata.obs[key_added]
     logging.info("Leiden results:\n%s", lbs.value_counts())
     return lbs
+
+
+def _augment_full_repeat(
+        x: np.ndarray,
+        y: np.ndarray = None,
+        n_add: int = 10,
+        seed: int = 1234,
+):
+    # This function do NOT consider the groups labeled in y
+    n0 = x.shape[0]
+    np.random.seed(seed)
+    ids_repeat = np.random.choice(n0, size=n_add, replace=True)
+    if sparse.issparse(x):
+        x_pseudo = x[ids_repeat,]
+    else:
+        x_pseudo = np.take(x, ids_repeat, axis=0)
+    y_pseudo = np.take(y, ids_repeat, axis=0) if y is not None else None
+    return x_pseudo, y_pseudo
+
+
+def _augment_balance_group_repeat(
+        x: np.ndarray,
+        y: np.ndarray,
+        n_tot_each: int = 1000,
+        groups: Optional[Sequence] = None,
+        concat: bool = True,
+        seed: int = 1234,
+):
+    # TODO: in case that y has multi-dimensions
+    # groups = np.unique(y, axis=0)
+    groups = set(groups) if groups else set(y)
+    groups = sorted(groups)
+    x_is_sparse = sparse.issparse(x)
+    if not x_is_sparse:
+        x = np.asarray(x)
+    y = np.asarray(y)
+    x_pseudo_all, y_pseudo_all = [], []
+    for lb in groups:
+        ids = np.flatnonzero(y == lb)
+        _n_add = n_tot_each - len(ids)
+        if _n_add <= 0:
+            continue
+        else:
+            logging.debug(lb)
+            _x = x[ids,]
+            _y = np.take(y, ids, axis=0)
+            _x_pseudo, _y_pseudo = _augment_full_repeat(_x, _y, n_add=_n_add,
+                                                        seed=seed)
+        x_pseudo_all.append(_x_pseudo)
+        y_pseudo_all.append(_y_pseudo)
+    if len(x_pseudo_all) == 1:
+        logging.warning(f"Only one group {groups[0]} was found!")
+        x_pseudo_all = x_pseudo_all[0]
+        y_pseudo_all = y_pseudo_all[0]
+    else:
+        if x_is_sparse:
+            x_pseudo_all = sparse.vstack(x_pseudo_all)
+        else:
+            x_pseudo_all = np.concatenate(x_pseudo_all, axis=0)
+        y_pseudo_all = np.concatenate(y_pseudo_all, axis=0)
+    if concat:
+        x_aug, y_aug, is_pseudo = _concat_x_and_y(
+            [x, y], [x_pseudo_all, y_pseudo_all])
+        is_pseudo = is_pseudo.astype(bool)
+        return x_aug, y_aug, is_pseudo
+    else:
+        return x_pseudo_all, y_pseudo_all
+
+
+def _concat_x_and_y(xy1, xy2):
+    x1, y1 = xy1
+    x2, y2 = xy2
+    n1, n2 = x1.shape[0], x2.shape[0]
+    if sparse.issparse(x1):
+        x = sparse.vstack([x1, x2])
+    else:
+        x = np.concatenate([x1, x2], axis=0)
+    y = np.concatenate([y1, y2], axis=0)
+    is_pseudo = np.array([False] * n1 + [True] * n2)
+    return x, y, is_pseudo
+
+
+def augment_repeat_adata(
+        adata: sc.AnnData,
+        key_y: str or int,  # column name in adata.obs
+        n_tot_each: int = 1000,
+        groups: Optional[Sequence] = None,
+        concat: bool = True,
+        seed: int = 1234,
+        id_prefix: str = 'aug',
+):
+    x = adata.X
+    y = adata.obs[key_y].values  # group labels
+    key_pseudo = 'is_pseudo'
+    x_pseudo, y_pseudo = _augment_balance_group_repeat(x, y, n_tot_each,
+                                                       groups=groups,
+                                                       concat=False, seed=seed)
+    ids_pseudo = [f'{id_prefix}.{i}' for i in range(x_pseudo.shape[0])]
+    obs_pseudo = pd.DataFrame({
+        key_y: y_pseudo, key_pseudo: True,
+    }, index=ids_pseudo)
+
+    if concat:
+        adata.obs[key_pseudo] = False
+        return sc.AnnData(
+            sparse.vstack([x, x_pseudo]),
+            obs=pd.concat([adata.obs, obs_pseudo], axis=0),
+            var=adata.var, )
+    else:
+        return sc.AnnData(x_pseudo, obs=obs_pseudo, var=adata.var, )
+
+
+def __test__():
+    fp = '/Users/yanyan/PycharmProjects/CAME/came/sample_data/raw-Baron_mouse.h5ad'
+    adt = sc.read_h5ad(fp)
+    key = 'cell_ontology_class'
+    print(adt.obs[key].value_counts())
+    adt = augment_repeat_adata(
+        adt, key_y=key, n_tot_each=100,
+        # groups=None,
+        groups=['T cell'],
+    )
+    print(adt.obs[key].value_counts())
+    print(adt.obs['is_pseudo'].value_counts())
