@@ -4,6 +4,7 @@ Created on Sun Apr 11 19:15:45 2021
 
 @author: Xingyan Liu
 """
+import logging
 from pathlib import Path
 import os
 from typing import Sequence, Union, Mapping, Optional, List
@@ -65,6 +66,7 @@ def prepare4train(
         key_class,
         key_clust='clust_lbs',
         scale_within=True,
+        batch_keys=None,
         unit_var=True,
         clip=False,
         clip_range=(-3, 3.5),
@@ -78,14 +80,28 @@ def prepare4train(
 ) -> dict:
     """
     dpair: DataPair
-    test_idx: if provided, should be an index-sequence of the same length as 
+    batch_keys:
+        a list of two strings (or None), specifying the batch-keys for
+        data1 and data2, respectively.
+        If given, features (of cell nodes) will be scaled within each batch
+         e.g., ['batch', 'sample']
+    test_idx:
+        By default, the testing indices will be decided automatically.
+        if provided, should be an index-sequence of the same length as
         `cluster_labels`
     """
     if key_clust and cluster_labels is None:
-        cluster_labels = dpair.obs_dfs[1][key_clust].values
+        try:
+            cluster_labels = dpair.obs_dfs[1][key_clust].values
+        except KeyError:
+            logging.warning(
+                f"`cluster_labels` is None and `key_clust={key_clust}` is NOT"
+                f"found in `dpair.obs_dfs[1].columns`, so not cluster labels"
+                f"will be adopted!")
 
     feat_dict = dpair.get_feature_dict(scale=scale_within,
                                        unit_var=unit_var,
+                                       batch_keys=batch_keys,
                                        clip=clip,
                                        clip_range=clip_range)
     labels, classes = dpair.get_obs_labels(
@@ -147,7 +163,7 @@ class Trainer(BaseTrainer):
             cluster_labels=cluster_labels,
             lr=lr,
             l2norm=l2norm,  # 1e-2 is tested for all datasets
-            use_cuda=use_cuda,
+            # use_cuda=use_cuda,
             dir_main=dir_main,
             **kwds  # for code compatibility (not raising error)
         )
@@ -213,10 +229,10 @@ class Trainer(BaseTrainer):
               use_class_weights=True,
               params_lossfunc={},
               n_pass=100,
-              eps=1e-4,
+              eps=5e-3,
               cat_class='cell',
               device=None,
-              backup_stride: int = 43,
+              backup_stride: int = 111,
               **other_inputs):
         """ Main function for model training (whole-graph based)
 
@@ -254,7 +270,8 @@ class Trainer(BaseTrainer):
             self.all_to_device(device)
         # train_labels_1hot = self.train_labels_1hot.to(device)
         train_labels_1hot = onehot_encode(
-            self.train_labels, self.classes, sparse_output=False, astensor=True).to(device)
+            self.train_labels, self.classes,
+            sparse_output=False, astensor=True).to(device)
 
         if use_class_weights:
             class_weights = to_device(self.class_weights, device)
@@ -274,20 +291,20 @@ class Trainer(BaseTrainer):
             self.optimizer.zero_grad()
             t0 = time.time()
             logits = self.model(feat_dict, self.g, **other_inputs)[cat_class]
-            logits2 = self.model(feat_dict, self.g, **other_inputs)[cat_class]
-            loss = ce_loss_with_rdrop(
-                logits, logits2, labels=train_labels,
-                labels_1hot=train_labels_1hot,
-                train_idx=train_idx, weight=class_weights,
-                loss_fn=classification_loss,
-                **params_lossfunc
-            )
-            # loss = classification_loss(
-            #     logits[train_idx],
-            #     train_labels, labels_1hot=train_labels_1hot,
-            #     weight=class_weights,
+            # logits2 = self.model(feat_dict, self.g, **other_inputs)[cat_class]
+            # loss = ce_loss_with_rdrop(
+            #     logits, logits2, labels=train_labels,
+            #     labels_1hot=train_labels_1hot,
+            #     train_idx=train_idx, weight=class_weights,
+            #     loss_fn=classification_loss,
             #     **params_lossfunc
             # )
+            loss = classification_loss(
+                logits[train_idx], train_labels,
+                labels_1hot=train_labels_1hot,
+                weight=class_weights,
+                **params_lossfunc
+            )
 
             # prediction 
             _, y_pred = torch.max(logits, dim=1)
@@ -315,10 +332,14 @@ class Trainer(BaseTrainer):
             if self._cur_epoch >= n_pass - 1:
                 self.ami_max = max(self.ami_max, ami)
                 if ami >= self.ami_max - eps > 0:
-                    self._cur_epoch_best = self._cur_epoch
+                    if ami >= self.ami_max:
+                        self._cur_epoch_best = self._cur_epoch
+                        _flag = 'current best'
+                    else:
+                        _flag = 'potential best'
                     self.save_model_weights()
                     backup = True
-                    print('[current best] model weights backup')
+                    print(f'[{_flag}] model weights backup')
                 elif self._cur_epoch % backup_stride == 0:
                     self.save_model_weights()
                     backup = True
@@ -369,7 +390,7 @@ class Trainer(BaseTrainer):
                         batch_size=128,
                         sampler=None,
                         device=None,
-                        backup_stride: int = 43,
+                        backup_stride: int = 111,
                         mod_info: int = 3,
                         **other_inputs):
         """ Main function for model training (based on mini-batches)
@@ -426,31 +447,26 @@ class Trainer(BaseTrainer):
                 # mfgs = [blk.to(device) for blk in mfgs]
 
                 logits = model(_feat_dict, mfgs, **other_inputs)[cat_class]
-                logits2 = model(_feat_dict, mfgs, **other_inputs)[cat_class]
+                # logits2 = model(_feat_dict, mfgs, **other_inputs)[cat_class]
 
                 # out_train_labels
                 out_train_lbs1hot = onehot_encode(
                     _labels, classes=self.classes, astensor=True).to(device)
                 _labels = to_device(_labels, device)
 
-                # out_train_lbs1hot = to_device(
-                #     self.train_labels_1hot[output_nodes[cat_class]].clone().detach(),
-                #     device,)
-
-                loss = ce_loss_with_rdrop(
-                    logits, logits2, labels=_labels,
-                    labels_1hot=out_train_lbs1hot,
-                    train_idx=None, weight=class_weights,
-                    loss_fn=classification_loss,
-                    **params_lossfunc
-                )
-                # loss = classification_loss(
-                #     logits[batch_train_idx],
-                #     to_device(out_train_labels, device),
-                #     labels_1hot=to_device(out_train_lbs1hot, device),
-                #     weight=class_weights,
+                # loss = ce_loss_with_rdrop(
+                #     logits, logits2, labels=_labels,
+                #     labels_1hot=out_train_lbs1hot,
+                #     train_idx=None, weight=class_weights,
+                #     loss_fn=classification_loss,
                 #     **params_lossfunc
                 # )
+                loss = classification_loss(
+                    logits, _labels,
+                    labels_1hot=out_train_lbs1hot,
+                    weight=class_weights,
+                    **params_lossfunc
+                )
                 loss.backward()
                 self.optimizer.step()
 
@@ -464,6 +480,7 @@ class Trainer(BaseTrainer):
             all_test_preds = []
             test_labels = []
             with tqdm.tqdm(test_dataloader) as tq, torch.no_grad():
+                model.eval()
                 for input_nodes, output_nodes, mfgs in tq:
                     inputs = to_device(idx_hetero(self.feat_dict, input_nodes), device)
                     mfgs = to_device(mfgs, device)
@@ -475,7 +492,7 @@ class Trainer(BaseTrainer):
                     )
                 all_test_preds = torch.cat(all_test_preds, dim=0)
                 test_labels = torch.cat(test_labels, dim=0)
-
+            model.train()
             # ========== evaluation (Acc.) ==========
             with torch.no_grad():
                 train_acc = accuracy(train_labels, all_train_preds)
@@ -495,11 +512,15 @@ class Trainer(BaseTrainer):
                 backup = False
                 if self._cur_epoch >= n_pass - 1:
                     self.ami_max = max(self.ami_max, ami)
-                    if ami > self.ami_max - eps > 0:
-                        self._cur_epoch_best = self._cur_epoch
+                    if ami >= self.ami_max - eps > 0:
+                        if ami >= self.ami_max:
+                            self._cur_epoch_best = self._cur_epoch
+                            _flag = 'current best'
+                        else:
+                            _flag = 'potential best'
                         self.save_model_weights()
                         backup = True
-                        print('[current best] model weights backup')
+                        print(f'[{_flag}] model weights backup')
                     elif self._cur_epoch % backup_stride == 0:
                         self.save_model_weights()
                         backup = True
